import { Connection } from 'typeorm';
import { InjectConnection } from '@nestjs/typeorm';
import { Injectable, ConsoleLogger } from '@nestjs/common';
import { ImageService } from './image/image.service';
import { AvatarApiService } from './avatar-api/avatar-api.service';
import { Avatar } from './entities/Avatar.entity';

@Injectable()
export class AppService extends ConsoleLogger {
  constructor(
    @InjectConnection()
    private db: Connection,
    private image: ImageService,
    private avatarApi: AvatarApiService,
  ) {
    super('app');
  }

  private async saveAvatar(url: string, size: number, buffer: Buffer) {
    const avatar = new Avatar();
    avatar.url = url;
    avatar.size = size;
    avatar.fromBuffer(buffer);
    return this.db.getRepository(Avatar).save(avatar);
  }

  async getImage(username: string, size: number) {
    const url = await this.avatarApi.getURLFromUsername(username);
    const existingAvatar = await this.db.getRepository(Avatar).findOne({
      where: {
        url,
        size,
      },
      select: ['content'],
    });
    if (existingAvatar) {
      return existingAvatar.toBuffer();
    }
    this.log(`Resizing ${url} to ${size}`);
    let originalAvatarBuffer: Buffer;
    const originalAvatar = await this.db.getRepository(Avatar).findOne({
      where: {
        url,
        size: 0,
      },
      select: ['content'],
    });
    if (originalAvatar) {
      originalAvatarBuffer = originalAvatar.toBuffer();
    } else {
      originalAvatarBuffer = await this.avatarApi.downloadAvatar(url);
      try {
        await this.saveAvatar(url, 0, originalAvatarBuffer);
      } catch (e) {
        this.error(`Fail to save original avatar: ${e.message}`);
      }
    }
    const avatarBuffer = await this.image.resize(originalAvatarBuffer, size);
    try {
      await this.saveAvatar(url, size, avatarBuffer);
    } catch (e) {
      this.error(`Fail to save avatar: ${e.message}`);
    }
    return avatarBuffer;
  }
}
