import { ConsoleLogger, Injectable } from '@nestjs/common';
import { HttpService } from '@nestjs/axios';
import { lastValueFrom } from 'rxjs';
import _ from 'lodash';
import { BlankReturnMessageDto } from '../dto/ReturnMessage.dto';

@Injectable()
export class AvatarApiService extends ConsoleLogger {
  constructor(private readonly http: HttpService) {
    super('AvatarApiService');
  }

  async getRealUsername(originalUsername: string) {
    try {
      const {
        data: {
          user: { name, username },
        },
      } = await lastValueFrom(
        this.http.get(
          `https://ygobbs.com/users/${encodeURIComponent(
            originalUsername,
          )}.json`,
          {
            responseType: 'json',
          },
        ),
      );
      return _.uniq([originalUsername, username, name]);
    } catch (e) {
      this.error(`User ${originalUsername} not found: ${e.toString()}`);
      return [originalUsername];
    }
  }

  async getAvatarURL(usernames: string[]): Promise<string> {
    for (const username of usernames) {
      try {
        const {
          data: {
            user: { avatar: _avatar },
          },
        } = await lastValueFrom(
          this.http.get(
            `https://api.moecube.com/accounts/users/${encodeURIComponent(
              username,
            )}.json`,
            {
              responseType: 'json',
            },
          ),
        );
        const avatar = _avatar as string;
        if (
          avatar?.match(/^http(s)?:\/\//) &&
          !avatar.endsWith('default_avatar.jpg')
        ) {
          return avatar;
        }
      } catch (e) {
        this.error(`Avatar of ${username} not found: ${e.toString()}`);
      }
    }
    return 'https://cdn02.moecube.com:444/accounts/default_avatar.jpg';
  }

  async getURLFromUsername(username: string) {
    const possibleNames = await this.getRealUsername(username);
    return this.getAvatarURL(possibleNames);
  }

  async downloadAvatar(url: string) {
    try {
      const { data } = await lastValueFrom(
        this.http.get<Buffer>(url, {
          responseType: 'arraybuffer',
        }),
      );
      return data;
    } catch (e) {
      this.error(`Failed to download avatar from ${url}: ${e.toString()}`);
      throw new BlankReturnMessageDto(404, 'Failed to download avatar');
    }
  }
}
